# SPDX-License-Identifier: MIT
"""Countback

Track a countback of places for ordering points tallies.
The zeroth place is reserved for stage races where 1st
place on the stage finish is considered more significant
than 1st place in intermediates.

Example:

    >>> a = countback()
    >>> a[2] = 1  # 1 2nd place
    >>> a[3] += 3  # 3 3rd places
    >>> str(a)
    '-,-,1,3'
    >>> b = countback('-,1,-,1')
    >>> b > a
    True

To sort a tally with countback, negate the countback
and points:

    st = []
    for rno, pts, cb in something:
        st.append((-pts, -cb, strops.bibstr_key(rno)))
    st.sort()


"""


class countback:
    """Dict wrapper for countback load/store/compare."""
    __hash__ = None

    def maxplace(self):
        """Return maximum place with non-zero count."""
        ret = 0
        if len(self.__store) > 0:
            for k, v in self.__store.items():
                if v != 0 and k > ret:
                    ret = k
        return ret

    def __init__(self, cbstr=None):
        self.__store = {}
        if cbstr is not None:
            self._fromstring(cbstr)

    def _fromstring(self, cbstr):
        """Re-populate store with counts from provided cb string."""
        propmap = {}
        cbvec = cbstr.split(',')
        if len(cbvec) > 0:
            for i, v in enumerate(cbvec):
                v = v.strip()
                if v and v != '-':
                    propmap[i] = int(v)
        # update store after entire string has been processed
        self.__store = {}
        for k, v in propmap.items():
            self.__store[k] = v

    def items(self):
        return self.__store.items()

    def keys(self):
        return self.__store.keys()

    def __repr__(self):
        return 'countback({})'.format(repr(str(self)))

    def __str__(self):
        ret = []
        for i in range(0, self.maxplace() + 1):
            v = self[i]
            if v != 0:
                ret.append(str(v))
            else:
                ret.append('-')
        return ','.join(ret)

    def __getitem__(self, key):
        # Note: unlike defaultdict, a missing key should not be
        #       added to store until it has a non-default value
        if key in self.__store:
            return self.__store[key]
        else:
            return 0

    def __setitem__(self, key, value):
        if isinstance(key, int) and key >= 0:
            self.__store[key] = value
        else:
            raise AttributeError('Countback keys must be integer >= 0')

    def __delitem__(self, key):
        del (self.__store[key])

    def __iter__(self):
        return iter(self.__store.keys())

    def __contains__(self, item):
        return item in self.__store

    def __lt__(self, other):
        if not isinstance(other, countback):
            return NotImplemented
        ret = False  # assume all same
        for i in range(0, max(self.maxplace(), other.maxplace()) + 1):
            a = self[i]
            b = other[i]
            if a != b:
                ret = a < b
                break
        return ret

    def __le__(self, other):
        if not isinstance(other, countback):
            return NotImplemented
        ret = True  # assume all same
        for i in range(0, max(self.maxplace(), other.maxplace()) + 1):
            a = self[i]
            b = other[i]
            if a != b:
                ret = a < b
                break
        return ret

    def __eq__(self, other):
        if not isinstance(other, countback):
            return NotImplemented
        ret = True
        for i in range(0, max(self.maxplace(), other.maxplace()) + 1):
            if self[i] != other[i]:
                ret = False
                break
        return ret

    def __ne__(self, other):
        if not isinstance(other, countback):
            return NotImplemented
        ret = False
        for i in range(0, max(self.maxplace(), other.maxplace()) + 1):
            if self[i] != other[i]:
                ret = True
                break
        return ret

    def __gt__(self, other):
        if not isinstance(other, countback):
            return NotImplemented
        ret = False  # assume all same
        for i in range(0, max(self.maxplace(), other.maxplace()) + 1):
            a = self[i]
            b = other[i]
            if a != b:
                ret = a > b
                break
        return ret

    def __ge__(self, other):
        if not isinstance(other, countback):
            return NotImplemented
        ret = True  # assume all same
        for i in range(0, max(self.maxplace(), other.maxplace()) + 1):
            a = self[i]
            b = other[i]
            if a != b:
                ret = a > b
                break
        return ret

    def __add__(self, other):
        if not isinstance(other, countback):
            return NotImplemented
        ret = countback()
        for i in range(0, max(self.maxplace(), other.maxplace()) + 1):
            v = self[i] + other[i]
            if v != 0:
                ret[i] = v
        return ret

    def __sub__(self, other):
        if not isinstance(other, countback):
            return NotImplemented
        ret = countback()
        for i in range(0, max(self.maxplace(), other.maxplace()) + 1):
            v = self[i] - other[i]
            if v != 0:
                ret[i] = v
        return ret

    def __neg__(self):
        ret = countback()
        for i in range(0, self.maxplace() + 1):
            v = self[i]
            if v != 0:
                ret[i] = -v
        return ret



# SPDX-License-Identifier: MIT
"""Read weather data from Comet device or AC Weather API"""
import logging
import requests
import threading
import json
import os
from random import random
from contextlib import suppress
from time import sleep, time
from metarace import sysconf, savefile, default_file
from metarace.tod import mkagg
from urllib3.util import Url
from uuid import UUID

# Configuration defaults - via AC Weather API
_SOURCE = 'ac'
_HOSTNAME = 'weather.auscycling.org.au'
_PORT = 443
_USETLS = True
_LOCATION_ENDPOINT = '/api/locations'
_DATA_ENDPOINT = '/api/data'
_POLLTIME = 59
_TIMEOUT = 10  # request timeout
_LOCATIONSCACHE = 'locations.json'
_ACSTALE = 900  # ~15 minutes validity

# Logging
_log = logging.getLogger('weather')
_log.setLevel(logging.DEBUG)
logging.getLogger('urllib3').setLevel(logging.ERROR)

_CONFIG_SCHEMA = {
    'ttype': {
        'prompt': 'Weather Data',
        'control': 'section',
    },
    'source': {
        'attr': 'source',
        'prompt': 'Source:',
        'control': 'choice',
        'hint': 'Data source',
        'defer': True,
        'conflicts': {
            'locationep': {
                'enable': 'ac',
                'disable': 'comet',
            },
            'usetls': {
                'enable': 'comet',
                'disable': 'ac',
            },
        },
        'options': {
            'ac': 'AusCycling Weather API',
            'comet': 'Comet Weather Station',
        },
        'default': 'ac',
    },
    'facility': {
        'prompt': 'Facility:',
        'attr': 'facility',
        'hint': 'Default facility code',
        'control': 'short',
        'default': None,
        'defer': True,
    },
    'dataep': {
        'prompt': 'Data:',
        'attr': 'dataep',
        'hint': 'Endpoint for weather data readings',
        'default': _DATA_ENDPOINT,
        'defer': True,
    },
    'locationep': {
        'prompt': 'Locations:',
        'attr': 'locationep',
        'hint': 'Endpoint for location IDs',
        'default': _LOCATION_ENDPOINT,
        'defer': True,
    },
    'hostname': {
        'prompt': 'Hostname:',
        'attr': 'hostname',
        'hint': 'Hostname of weather service',
        'default': _HOSTNAME,
        'defer': True,
    },
    'port': {
        'prompt': 'Port:',
        'attr': 'port',
        'control': 'short',
        'hint': 'TCP port number',
        'type': 'int',
        'default': _PORT,
        'defer': True,
    },
    'usetls': {
        'prompt': 'Security:',
        'attr': 'usetls',
        'subtext': 'Use TLS?',
        'type': 'bool',
        'control': 'check',
        'hint': 'Connect using TLS',
        'default': _USETLS,
        'defer': True,
    },
    'polltime': {
        'prompt': 'Interval:',
        'attr': 'polltime',
        'control': 'short',
        'subext': 'seconds',
        'hint': 'Delay in seconds between readings',
        'type': 'int',
        'default': _POLLTIME,
        'defer': True,
    },
    'timeout': {
        'prompt': 'Timeout:',
        'attr': 'timeout',
        'control': 'short',
        'subext': 'seconds',
        'hint': 'Request timeout in seconds',
        'type': 'int',
        'default': _TIMEOUT,
        'defer': True,
    },
}


def _getFloatKey(data, key, default=None, min=0, max=1):
    ret = default
    if key in data and isinstance(data[key], float):
        ret = data[key]
        if ret is not None and ret < min:
            ret = None
        if ret is not None and ret > max:
            ret = None
    return ret


def _readAcWeather(data):
    """Read temperature, pressure and humidity from AC Weather record"""
    ret = None
    t = None
    h = None
    p = None
    if isinstance(data, dict):
        lr = _getFloatKey(data,
                          'lastUpdated',
                          default=_ACSTALE,
                          min=0,
                          max=86400)
        t = _getFloatKey(data, 'temperature', min=-30, max=50)
        h = _getFloatKey(data, 'humidity', min=0, max=120)
        p = _getFloatKey(data, 'pressure', min=500, max=1100)
    if t is not None and h is not None and p is not None and lr is not None:
        if lr < _ACSTALE:
            ret = (t, h, p)
        else:
            lt = mkagg(str(lr))
            _log.info('AC Weather stale %s: %0.1f\u00B0C, %0.1f%%, %0.1fhPa',
                      lt.rawtime(0), t, h, p)
    else:
        _log.info('AC Weather invalid: %r, %r, %r, %r', t, h, p, lr)
    return ret


def _readComet(channels):
    """Read temperature, pressure and humidity from comet json channels"""
    # Assumes channel setup:
    #       1: Temperature (degrees C)
    #       2: Relative Humidity (%RH)
    #       5: Barometric Pressure (hPa)
    ret = None
    t = None
    h = None
    p = None
    if isinstance(channels, list):
        with suppress(TypeError, ValueError):
            for channel in channels:
                if channel['name'] == 'Temperature':
                    t = float(channel['value'])
                elif channel['name'] == 'Relative humidity':
                    h = float(channel['value'])
                elif channel['name'] == 'Barometric pressure':
                    p = float(channel['value'])
    if t is not None and h is not None and p is not None:
        ret = (t, h, p)
    return ret


def Weather(facility=None):
    """Return a weather object based on configured source"""
    sysconf.add_section('weather', _CONFIG_SCHEMA)
    if sysconf.get_value('weather', 'source') == 'ac':
        return ACWeather(facility)
    else:
        if facility is None:
            facility = 'comet'
        return Comet(facility)


class BaseWeather(threading.Thread):
    """Base class for weather observations."""

    def __init__(self, facility=None):
        """Construct weather thread object."""
        threading.Thread.__init__(self, daemon=True)
        self.t = 0.0
        self.h = 0.0
        self.p = 0.0
        self._lt = None
        self._running = None
        self._s = None

        sysconf.add_section('weather', _CONFIG_SCHEMA)
        self._locationid = None
        self._facility = facility
        if self._facility is None:
            self._facility = sysconf.get_value('weather', 'facility')
        self._hostname = sysconf.get_value('weather', 'hostname')
        self._port = sysconf.get_value('weather', 'port')
        self._usetls = sysconf.get_value('weather', 'usetls')
        self._polltime = sysconf.get_value('weather', 'polltime')
        self._timeout = sysconf.get_value('weather', 'timeout')
        self._dataep = sysconf.get_value('weather', 'dataep')
        self._locationep = sysconf.get_value('weather', 'locationep')
        # todo: strip relative path from front of endpoints
        self._scheme = 'https' if self._usetls else 'http'

        # sanity check params before spinning up
        if not self._facility or not self._hostname:
            _log.debug('Weather service not configured')
            self._running = False

    def valid(self):
        """Return true if current measurements are valid."""
        age = None
        if self._lt is not None:
            # compare with now and return age
            age = time() - self._lt
        return age is not None and age < 2 * self._polltime

    def exit(self):
        """Request thread termination."""
        self._running = False

    def _read(self):
        """Update weather readings"""
        pass

    def run(self):
        """Called via Thread.start()."""
        _log.debug('Starting %s[%s]', self._facility, self.native_id)
        if self._running is None:
            # may be set False before run is called
            self._running = True
        with requests.Session() as self._s:
            while self._running:
                try:
                    self._read()
                except Exception as e:
                    _log.warning('%s[%s] %s: %s', self._facility,
                                 self.native_id, e.__class__.__name__, e)
                sleep(self._polltime + 0.25 * self._polltime * random())
        _log.debug('Exit %s[%s]', self._facility, self.native_id)

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.exit()
        if exc_type is not None:
            return False  # raise exception
        self.join()
        return True


class ACWeather(BaseWeather):
    """AC Weather API object class."""

    def set_facility(facility):
        """Update the facility code."""
        self._lt = None
        self._locationid = None
        self._facility = facility

    def _updatelocations(self):
        """Fetch the locations data from AC Weather."""
        try:
            url = Url(scheme=self._scheme,
                      host=self._hostname,
                      port=self._port,
                      path=self._locationep).url
            r = self._s.get(url, timeout=self._timeout)
            if r.status_code == 200:
                jd = r.json()
                with savefile(_LOCATIONSCACHE) as f:
                    json.dump(jd, f, indent=1)
                _log.debug('Loaded %d AC Weather locations from %s', len(jd),
                           url)
            else:
                _log.debug('Invalid location response: %d', r.status_code)
        except Exception as e:
            _log.debug('%s loading AC Weather locations: %s',
                       e.__class__.__name__, e)

    def _getlocation(self):
        """Fetch the location id for the configured facility code."""
        if self._locationid is not None:
            return self._locationid
        lcache = default_file(_LOCATIONSCACHE)
        if not os.path.exists(lcache):
            self._updatelocations()
        if os.path.exists(lcache):
            with open(lcache) as f:
                for v in json.load(f):
                    if 'venueCode' in v and v['venueCode'] == self._facility:
                        if 'id' in v and v['id']:
                            self._locationid = str(UUID(v['id']))
                            break
                        else:
                            _log.debug('AC Weather invalid location')
                else:
                    _log.debug('AC Weather facility %s not found',
                               self._facility)
        else:
            _log.debug('AC Weather locations cache file missing')
        return self._locationid

    def _read(self):
        """Update weather readings."""
        if self._getlocation() is None:
            _log.warning('Location ID for %s not found, weather disabled',
                         self._facility)
            self._running = False
            return
        try:
            url = Url(scheme=self._scheme,
                      host=self._hostname,
                      port=self._port,
                      path=self._dataep).url
            r = self._s.get(url,
                            params={'locationId': self._locationid},
                            timeout=self._timeout)
            if r.status_code == 200:
                nv = _readAcWeather(r.json())
                if nv is not None:
                    self._lt = time()
                    self.t = nv[0]
                    self.h = nv[1]
                    self.p = nv[2]
                    _log.debug('%s: %0.1f\u00B0C, %0.1f%%, %0.1fhPa',
                               self._facility, self.t, self.h, self.p)
                else:
                    _log.info('Error reading %s data from %s', self._facility,
                              self._hostname)
            else:
                _log.info('Invalid response for %s: %d', self._facility,
                          r.status_code)
        except Exception as e:
            _log.warning('%s reading %s weather from %s: %s',
                         e.__class__.__name__, self._facility, self._hostname,
                         e)


class Comet(BaseWeather):
    """Comet device object class."""

    def _read(self):
        """Fetch values from comet and update internal state"""
        try:
            url = Url(scheme=self._scheme,
                      host=self._hostname,
                      port=self._port,
                      path=self._dataep).url
            r = self._s.get(url, timeout=self._timeout)
            if r.status_code == 200:
                jd = r.json()
                if 'ch' in jd:
                    nv = _readComet(jd['ch'])
                    if nv is not None:
                        self._lt = time()  # use current host time
                        self.t = nv[0]
                        self.h = nv[1]
                        self.p = nv[2]
                        _log.debug('%s: %0.1f\u00B0C, %0.1f%%, %0.1fhPa',
                                   self._facility, self.t, self.h, self.p)
                    else:
                        _log.info('Error reading %s data from %s',
                                  self._facility, self._hostname)
                else:
                    _log.info('Invalid %s data from %s', self._facility,
                              self._hostname)
            else:
                _log.info('Invalid response for %s: %d', self._facility,
                          r.status_code)
        except Exception as e:
            _log.warning('%s reading %s weather from %s: %s',
                         e.__class__.__name__, self._facility, self._hostname,
                         e)



# SPDX-License-Identifier: MIT
"""String filtering, truncation and padding."""

import re
from random import randint
import grapheme
from contextlib import suppress

# replace codepoints 0->255 with space unless overridden
# "protective" against unencoded ascii strings and control chars
SPACEBLOCK = ''
for i in range(0, 256):
    SPACEBLOCK += chr(i)


# unicode translation 'map' class
class unicodetrans:

    def __init__(self, keep='', replace=SPACEBLOCK, replacechar=' '):
        self.comp = dict((ord(c), replacechar) for c in replace)
        for c in keep:
            self.comp[ord(c)] = c

    def __getitem__(self, k):  # override to return a None
        return self.comp.get(k)


INTEGER_UTRANS = unicodetrans('-0123456789')
NUMERIC_UTRANS = unicodetrans('-0123456789.e')
PLACELIST_UTRANS = unicodetrans(
    '-0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz')
PLACESERLIST_UTRANS = unicodetrans(
    '-.0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz')
BIBLIST_UTRANS = unicodetrans(
    '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz')
RIDERNO_UTRANS = unicodetrans(
    '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz', '', '')
BIBSERLIST_UTRANS = unicodetrans(
    '.0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz')
WEBFILE_UTRANS = unicodetrans(
    '_0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz', '.',
    '_')
# special case: map out controls, but keep everything else
PRINT_UTRANS = {}
for cp in range(0, 0x20):
    PRINT_UTRANS[cp] = ' '
for cp in range(0x7f, 0xa1):
    PRINT_UTRANS[cp] = ' '
PRINT_UTRANS[0x1680] = ' '
PRINT_UTRANS[0x180e] = ' '
PRINT_UTRANS[0x202f] = ' '
PRINT_UTRANS[0x205f] = ' '
PRINT_UTRANS[0x3000] = ' '
PRINT_UTRANS[0xffa0] = ' '

# timing channels - this duplicates defs in timy
CHAN_START = 0
CHAN_INT = 9
CHAN_UNKNOWN = -1

# running number comparisons
RUNNER_NOS = {
    'RED': 0,
    'WHI': 1,
    'BLU': 2,
    'YEL': 3,
    'GRN': 4,
    'PIN': 5,
    'BLA': 6,
    'GRY': 7,
    'ORA': 8,
    'PUR': 9,
    'RDW': 10,
    'BLW': 11,
    'YLW': 12,
    'GRW': 13
}

DNFCODEMAP = {
    'rel': -2,
    'otl': -1,
    'dnf': 1,
    'abd': 2,
    'dns': 3,
    'dsq': 4,
    '': 0
}
_UNPLACEDRANK = 100000  # Arbitrarily large ranking for unplaced


def rand_key(data=None):
    """Return a random integer key for shuffling."""
    return randint(0, 0xffffffff)


def riderno_key(bib):
    """Return a comparison key for sorting rider number strings."""
    return bibstr_key(bib)


def dnfcode_key(code):
    """Return a rank/dnf code sorting key."""
    ret = _UNPLACEDRANK
    if code is not None:
        code = code.lower()
        if code in DNFCODEMAP:
            ret += DNFCODEMAP[code]
        else:
            code = code.rstrip('.')
            with suppress(ValueError, TypeError):
                ret = int(code)
    return ret


def bibstr_key(bibstr=''):
    """Return a comparison key for sorting rider bib.ser strings."""
    bib, ser = bibstr2bibser(bibstr)
    return '% 6s:% 4s' % (ser, bib)


def randstr(data=None):
    """Return a string of random digits."""
    return str(randint(10000, 99999))


def promptstr(prompt='', value=''):
    """Prefix a non-empty string with a prompt, or return empty."""
    ret = ''
    if value:
        ret = prompt + ' ' + value
    return ret


def heatsplit(heatstr):
    """Return a sortable heat/lane pair for the supplied heat string."""
    hv = heatstr.split('.')
    while len(hv) < 2:
        hv.append('0')
    return (int(hv[0]), int(hv[1]))


def fitname(first, last, width, trunc=False):
    """Return a truncated name string of width or less graphemes
       shortened to fit with priority:

          Firstnamer FAMILY-LASTNAME
          Firstnamer LASTNAME
          F. FAMILY-LASTNAME
          F. LASTNAME
          LASTNAME
          F. LAST...
    """
    # Note: Use rider.fitname() for a caching version

    # Full name: 'Firstname FAMILY-LASTNAME'
    ret = ''
    fn = first.strip().title()
    fl = grapheme.length(fn)
    ln = last.strip().upper()
    ll = grapheme.length(ln)
    flen = fl + ll
    if fl and ll:
        flen += 1
    if flen > width:
        # Try without hyphen: 'Firstname LASTNAME'
        lshrt = ln.split('-')[-1].strip()
        lsl = grapheme.length(lshrt)
        flen = fl + lsl
        if fl and lsl:
            flen += 1
        if flen > width and ln:
            if fl > 2:
                # Retry with abbreviated firstname: 'F. FAMILY-LASTNAME'
                fshrt = grapheme.slice(fn, end=1) + '.'
                fn = fshrt
                fsl = 2
                flen = fsl + ll
                if fsl and ll:
                    flen += 1
        else:
            ln = lshrt
    if fn and ln:
        ret = ' '.join((fn, ln))
    elif fn:
        ret = fn
    else:
        ret = ln
    if trunc and grapheme.length(ret) > width:
        if width > 4:
            ret = grapheme.slice(ret, end=width - 1) + '\u2026'
        else:
            ret = grapheme.slice(ret, end=width)
    return ret


def drawno_encirc(drawstr=''):
    ret = ''
    with suppress(ValueError, TypeError):
        ival = int(drawstr)
        ret = drawstr
        if ival > 0 and ival <= 10:
            ret = (
                '\u00a0' +  # Hack: nbsp to get full line height
                chr(0x245f + ival))  # CP U+2460 "Circled digit"
    return ret


def rank2ord(place):
    """Return ordinal for the given place."""
    omap = {
        '1': 'st',
        '2': 'nd',
        '3': 'rd',
        '11': 'th',
        '12': 'th',
        '13': 'th'
    }
    ret = place
    if place.isdigit():
        if place in omap:
            ret = place + omap[place]
        elif len(place) > 1 and place[-2:] in omap:
            ret = place + omap[place[-2:]]
        else:
            if len(place) > 1 and place[-1] in omap:  # last digit 1,2,3
                ret = place + omap[place[-1]]
            else:
                ret = place + 'th'
    return ret


def rank2int(rank):
    """Convert a rank/placing string into an integer."""
    ret = None
    with suppress(ValueError, TypeError):
        ret = int(rank.replace('.', ''))
    return ret


def mark2mark(handicap):
    """Normalise a handicap string for use in trackmeet"""
    ret = ''
    mark = mark2int(str(handicap))
    if mark is not None:
        if mark == 0:
            ret = 'scr'
        else:
            ret = str(mark) + 'm'
    return ret


def mark2int(handicap):
    """Convert a handicap string into an integer number of metres."""
    handicap = handicap.strip().lower()
    ret = None  # not recognised as handicap
    if handicap != '':
        if handicap[0:3] == 'scr':  # 'scr{atch}'
            ret = 0
        else:  # try [number]m form
            with suppress(ValueError, TypeError):
                handicap = handicap.translate(INTEGER_UTRANS).strip()
                ret = int(handicap)
    return ret


def truncpad(srcline, width, align='l', ellipsis=True):
    """Return srcline truncated and padded to width aligned graphemes"""
    curlen = grapheme.length(srcline)
    if curlen > width:
        # truncation
        if ellipsis and width > 4:
            ret = grapheme.slice(srcline, end=width - 1) + '\u2026'
        else:
            ret = grapheme.slice(srcline, end=width)
    elif curlen < width:
        # padding
        residual = width - curlen
        if align == 'r':
            ret = ' ' * residual + srcline
        elif align == 'c':
            prelen = residual // 2
            postlen = residual - prelen
            ret = ' ' * prelen + srcline + ' ' * postlen
        else:
            ret = srcline + ' ' * residual
    else:
        ret = srcline
    return ret


def resname_bib(bib, first, last, club, series=''):
    """Return rider name formatted for results with bib."""
    bibstr = bibser2bibstr(bib, series)
    ret = [bibstr, ' ', fitname(first, last, 48)]
    if club is not None and club != '':
        if len(club) < 4:
            club = club.upper()
        ret.extend([' (', club, ')'])
    return ''.join(ret)


def resname(first, last=None, club=None):
    """Return rider name formatted for results."""
    ret = fitname(first, last, 48)
    if club is not None and club != '':
        if len(club) < 4:
            club = club.upper()
        ret = ''.join([ret, ' (', club, ')'])
    return ret


def listname(first, last=None, club=None):
    """Return a rider name summary field for non-edit lists."""
    ret = fitname(first, last, 32)
    if club:
        if len(club) < 4:
            club = club.upper()
        ret = ''.join([ret, ' (', club, ')'])
    return ret


def reformat_bibserlist(bibserstr):
    """Filter and return a bib.ser start list."""
    if not isinstance(bibserstr, str):
        return ''
    return ' '.join(bibserstr.translate(BIBSERLIST_UTRANS).split())


def reformat_bibserplacelist(placestr):
    """Filter and return a canonically formatted bib.ser place list."""
    if not isinstance(placestr, str):
        return ''
    if '-' not in placestr:  # This is the 'normal' case!
        return reformat_bibserlist(placestr)
    # otherwise, filter dead heats
    placestr = placestr.translate(PLACESERLIST_UTRANS).strip()
    placestr = re.sub(r'\s*\-\s*', r'-', placestr)  # remove surrounds
    placestr = re.sub(r'\-+', r'-', placestr)  # combine dupes
    return ' '.join(placestr.strip('-').split())


def reformat_biblist(bibstr):
    """Filter and return a canonically formatted start list."""
    if not isinstance(bibstr, str):
        return ''
    return ' '.join(bibstr.translate(BIBLIST_UTRANS).split())


def riderlist_split(riderstr, rdb=None, series=''):
    """Filter, search and return a list of matching riders for entry."""
    if not isinstance(riderstr, str):
        return []

    riderstr = riderstr.upper()
    dstlist = []

    # extract categories and 'all' from riderstr
    if rdb is not None:
        srclist = []
        catnames = rdb.listcats(series)
        for r in riderstr.split():
            if r == 'ALL':
                for rid in rdb.biblistfromseries(series):
                    bib, ser = bibstr2bibser(rid)
                    if ser == series:
                        dstlist.append(bib)
            elif r in catnames:
                for rid in rdb.biblistfromcat(r, series):
                    bib, ser = bibstr2bibser(rid)
                    if ser == series:
                        dstlist.append(bib)
            else:
                srclist.append(r)
        riderstr = ' '.join(srclist)

    # expand ranges in riderstr
    riderstr = reformat_placelist(riderstr)
    for nr in riderstr.split():
        if '-' in nr:
            # try for a range...
            l = None
            n = None
            for r in nr.split('-'):
                if l is not None:
                    if l.isdigit() and r.isdigit():
                        start = int(l)
                        end = int(r)
                        if start < end:
                            c = start
                            while c < end:
                                dstlist.append(str(c))
                                c += 1
                        else:
                            dstlist.append(l)
                    else:
                        # one or both not ints
                        dstlist.append(l)
                else:
                    pass
                l = r
            if l is not None:  # catch final value
                dstlist.append(l)
        else:
            dstlist.append(nr)

    ret = []
    rset = set()
    for r in dstlist:
        r = r.upper()
        if r not in rset:
            ret.append(r)
            rset.add(r)
    return ret


def placeset(spec=''):
    """Convert a place spec into an ordered set of place ints."""

    # NOTE: ordering of the set must be retained to correctly handle
    #       autospecs where the order of the places is not increasing
    #       eg: sprint semi -> sprint final, the auto spec is: 3,1,2,4
    #       so the 'winners' go to the gold final and the losers to the
    #       bronze final.
    ret = []
    spec = reformat_placelist(spec)
    # pass 1: expand ranges
    for nr in spec.split():
        if '-' in spec:
            # try for a range...
            l = None
            n = None
            for r in nr.split('-'):
                if l is not None:
                    if l.isdigit() and r.isdigit():
                        start = int(l)
                        end = int(r)
                        if start < end:
                            c = start
                            while c < end:
                                ret.append(str(c))
                                c += 1
                        else:
                            ret.append(l)  # give up on last val
                    else:
                        # one or both not ints
                        ret.append(l)
                else:
                    pass
                l = r
            if l is not None:  # catch final value
                ret.append(l)
        else:
            ret.append(nr)
    # pass 2: filter out non-numbers, only places considered
    rset = []
    for i in ret:
        if i.isdigit():
            ival = int(i)
            if ival not in rset:
                rset.append(ival)
    return rset


def reformat_placelist(placestr):
    """Filter and return a canonically formatted place list."""
    if '-' not in placestr:
        return reformat_biblist(placestr)
    placestr = placestr.translate(PLACELIST_UTRANS).strip()
    placestr = re.sub(r'\s*\-\s*', r'-', placestr)  # remove surrounds
    placestr = re.sub(r'\-+', r'-', placestr)  # combine dupes
    return ' '.join(placestr.strip('-').split())


def confopt_bool(confstr):
    """Check and return a boolean option from config."""
    if isinstance(confstr, str):
        if confstr.lower() in (
                'yes',
                'true',
                '1',
        ):
            return True
        else:
            return False
    else:
        return bool(confstr)


def plural(count=0):
    """Return plural extension for provided count."""
    ret = 's'
    if count == 1:
        ret = ''
    return ret


def confopt_str(confob, default=None):
    """Check and return a plain string for the provided value."""
    ret = default
    if isinstance(confob, str):
        ret = confob
    return ret


def confopt_riderno(confstr, default=''):
    """Check and return rider number, filtered only."""
    return confstr.translate(RIDERNO_UTRANS).strip()


def confopt_float(confstr, default=None):
    """Check and return a floating point number."""
    ret = default
    with suppress(ValueError, TypeError):
        ret = float(confstr)
    return ret


def confopt_posfloat(confstr, default=None):
    """Check and return a positive floating point number."""
    ret = default
    with suppress(ValueError, TypeError):
        nv = float(confstr)
        if nv > 0:
            ret = nv
    return ret


def confopt_distunits(confstr):
    """Check and return a valid unit from metres or laps."""
    if 'lap' in confstr.lower():
        return 'laps'
    else:
        return 'metres'


def confopt_int(confstr, default=None):
    """Check and return a valid integer."""
    ret = default
    with suppress(ValueError, TypeError):
        ret = int(confstr)
    return ret


def confopt_posint(confstr, default=None):
    """Check and return a valid positive integer."""
    ret = default
    with suppress(ValueError, TypeError):
        ret = int(confstr)
        if ret < 0:
            ret = default
    return ret


def confopt_dist(confstr, default=None):
    """Check and return a valid distance unit."""
    return confopt_posint(confstr, default)


def chan2id(chanstr='0'):
    """Return a channel ID for the provided string, without fail."""
    ret = CHAN_UNKNOWN
    with suppress(ValueError, TypeError):
        if isinstance(chanstr, str):
            chanstr = chanstr.upper().rstrip('M').lstrip('C')
            if chanstr.isdigit():
                ret = int(chanstr)
        else:
            ret = int(chanstr)
    if ret < CHAN_UNKNOWN or ret > CHAN_INT:
        ret = CHAN_UNKNOWN
    return ret


def id2chan(chanid=0):
    """Return a normalised channel string for the provided channel id."""
    ret = 'C?'
    if isinstance(chanid, int) and chanid >= CHAN_START and chanid <= CHAN_INT:
        ret = 'C' + str(chanid)
    return ret


def confopt_chan(confstr, default=None):
    """Check and return a valid timing channel id string."""
    ret = chan2id(confstr)
    if ret == CHAN_UNKNOWN:
        ret = chan2id(default)
    return ret


def confopt_pair(confstr, value, default=None):
    """Return value or the default."""
    ret = default
    if confstr.lower() == value.lower():
        ret = value
    return ret


def confopt_list(confstr, list=[], default=None):
    """Return an element from list or default."""
    ret = default
    look = confstr.lower()
    if look in list:
        ret = look
    return ret


def bibstr(bib=''):
    """Reformat and return competitor id"""
    return bibser2bibstr(*bibstr2bibser(bib))


def bibstr2bibser(bibstr=''):
    """Split a bib.series string and return bib and series."""
    a = bibstr.strip().split('.')
    ret_bib = ''
    ret_ser = ''
    if len(a) > 0:
        ret_bib = a[0].upper()
    if len(a) > 1:
        ret_ser = a[1].lower()
    return (ret_bib, ret_ser)


def bibser2bibstr(bib='', ser=''):
    """Return a competitor bib.series string."""
    ret = bib.upper()
    if ser != '':
        ret += '.' + ser.lower()
    return ret


def lapstring(lapcount=None):
    lapstr = ''
    if lapcount:
        lapstr = '%d\u2006Lap%s' % (
            lapcount,
            plural(lapcount),
        )
    return lapstr


def titlesplit(src='', linelen=24):
    """Split a string on word boundaries to try and fit into 3 fixed lines."""
    ret = ['', '', '']
    words = src.split()
    wlen = len(words)
    if wlen > 0:
        line = 0
        ret[line] = words.pop(0)
        for word in words:
            pos = len(ret[line])
            if pos + len(word) >= linelen:
                # new line
                line += 1
                if line > 2:
                    break
                ret[line] = word
            else:
                ret[line] += ' ' + word
    return ret



# SPDX-License-Identifier: MIT
"""Factors for combined categories"""
import logging
import requests
import decimal
import os
import csv
from contextlib import suppress
from metarace import sysconf, default_file, savefile
from metarace import tod
from tempfile import NamedTemporaryFile

# Configuration defaults
_MINFACTOR = decimal.Decimal('0.01')  # minimum allowed factor
_MAXFACTOR = decimal.Decimal(1)  # maximum allowed factor
_QUANT = decimal.Decimal('0.00001')  # desired factor precsion
_DIVISOR = 100000
_PLACES = 5
_ROUND = decimal.ROUND_HALF_EVEN
_FILENAME = 'factors'
_DEFAULT_FILE = _FILENAME + '.csv'
_DEFAULT_URL = None
_TIMEOUT = 10
_MISSING_FACTOR = _MAXFACTOR  # TBC
_CATEGORIES = {
    'PF': 'Para-cycling Factored',
    'OF': 'Open Factored',
    'MOF': 'Men Open Factored',
    'WOF': 'Women Open Factored',
    'AF': 'Age-Based Factored',
    'MAF': 'Men Age-Based Factored',
    'WAF': 'Women Age-Based Factored',
}

# Logging
_log = logging.getLogger('factors')
_log.setLevel(logging.DEBUG)
logging.getLogger('urllib3').setLevel(logging.ERROR)

_CONFIG_SCHEMA = {
    'ttype': {
        'prompt': 'Time Factors',
        'control': 'section',
    },
    'updateurl': {
        'prompt': 'Source:',
        'attr': 'updateurl',
        'hint': 'Filename or URL to update factors',
        'default': _DEFAULT_URL,
    },
}


def readFactor(factor, label, default=None, strict=True):
    """Check a provided factor and return Decimal value."""
    # Note: Also accepts float input, but will complain about quantisation
    ret = default
    try:
        mult = _DIVISOR
        floatval = float(factor)
        if floatval < _MINFACTOR:
            raise ValueError('%s=%s less than minimum %s' %
                             (label, factor, _MINFACTOR))
        if floatval > _MAXFACTOR:
            _log.info('%s=%s assumed percentage', label, factor)
            mult = 1000
        idv = round(mult * floatval)
        dcv = (decimal.Decimal(idv) / _DIVISOR).quantize(_QUANT,
                                                         rounding=_ROUND)
        if dcv > _MAXFACTOR:
            raise ValueError('%s=%s (%s) greater than maximum %s' %
                             (label, factor, dcv, _MAXFACTOR))
        dstr = str(dcv)
        if dstr != factor:
            if strict:
                if (dcv * _DIVISOR) == int(round(mult * floatval, 5)):
                    _log.info('%s=%s -> %s', label, factor, dstr)
                else:  # value has been rounded
                    raise ValueError('%s=%s -> %s' % (label, factor, dcv))
            else:  # only warn
                _log.warning('%s=%s -> %s', label, factor, dstr)
        ret = dcv
    except Exception as e:
        _log.warning('%s: %s', e.__class__.__name__, e)
    return ret


def _cleankey(keystr):
    if keystr:
        keystr = keystr.split(maxsplit=1)[0].upper()
    return keystr


class Factors:
    """Provide time factors from published source."""

    def __init__(self):
        self._valid = False
        self._store = {}
        for cat in _CATEGORIES:
            self._store[cat] = {}

    def is_valid(self):
        """Return True if currently loaded factors are valid."""
        return self._valid

    def factor_time(self, cat, sportclass, time, places=2, rounding=_ROUND):
        """Return a factored version of tod time."""
        ret = time
        try:
            factor = self.get_factor(cat, sportclass)
            nt = +time  # create local copy of time
            nt.timeval *= factor
            ret = nt.places(places, rounding, flag='FCTRD')
            ret.source = factor
            _log.debug('%s.%s: %s * %s -> %s', cat, sportclass, time.timeval,
                       factor, nt.timeval)
        except Exception as e:
            _log.warning('%s %s.%s: %s', e.__class__.__name__, cat, sportclass,
                         e)
        return ret

    def set_factor(self, cat, sportclass, factor):
        """Update a factor with the provided value."""
        cat = _cleankey(cat)
        sportclass = _cleankey(sportclass)
        label = '.'.join((cat, sportclass))
        fv = readFactor(factor, label)
        if fv is None:
            _log.warning('%s.%s invalid factor %s ignored', cat, sportclass,
                         factor)
            return False
        if cat not in self._store:
            _log.warning('%s invalid cat', cat)
            return False

        self._store[cat][sportclass] = fv
        self._valid = True
        _log.info('%s.%s updated factor %s', cat, sportclass, fv)
        return True

    def get_factor(self, cat, sportclass):
        """Return a factor for the named cat and sport class."""
        ret = _MISSING_FACTOR
        if self._valid:
            if cat in self._store:
                if sportclass in self._store[cat]:
                    ret = self._store[cat][sportclass]
                else:
                    _log.warning('%s.%s not found returning default', cat,
                                 sportclass)
            else:
                _log.warning('%s invalid cat, returning default', cat)
        else:
            _log.warning('Time Factors not valid')
        return ret

    def update(self, filename=None):
        """Fetch published factors, overwriting current values if valid."""
        if filename is None:
            filename = _DEFAULT_FILE
        sysconf.add_section('factors', _CONFIG_SCHEMA)
        url = sysconf.get_value('factors', 'updateurl')
        if not url:
            _log.error('Update URL not set, factors not updated')
            return self._valid
        try:
            _log.debug('Fetching updated factors from %s', url)
            with requests.Session() as s:
                r = s.get(url, timeout=_TIMEOUT)
                if r.status_code == 200:
                    # write data to temp file
                    with NamedTemporaryFile(mode='wb') as f:
                        f.write(r.content)
                        f.flush()
                        _log.debug('Temp factors saved to: %s', f.name)
                        with open(f.name) as g:
                            self.read(g)
                        if self._valid:
                            self.save(filename)
                else:
                    _log.error('Invalid read response: %d', r.status_code)
        except Exception as e:
            if self._valid:
                _log.warning('%s updating factors, values retained: %s',
                             e.__class__.__name__, e)
            else:
                _log.error('%s updating factors: %s', e.__class__.__name__, e)
        return self._valid

    def write(self, file):
        """Write factors to CSV file."""
        hdr = ('Category', 'Sport Class', 'Factor')
        cw = csv.DictWriter(file, fieldnames=hdr, quoting=csv.QUOTE_ALL)
        cw.writeheader()
        for cat, factors in self._store.items():
            _log.debug('Dumping %d factors for cat=%s', len(factors), cat)
            for sportclass, factor in factors.items():
                cw.writerow({
                    'Category': cat,
                    'Sport Class': sportclass,
                    'Factor': str(factor)
                })

    def save(self, filename=None):
        """Save current factors to filename if valid."""
        if not self.is_valid():
            _log.error('Factors not valid, not saved.')
            return self._valid
        if filename is None:
            filename = _DEFAULT_FILE
        try:
            with savefile(filename) as f:
                self.write(f)
        except Exception as e:
            _log.warning('%s save: %s', e.__class__.__name__, e)

    def read(self, file):
        count = 0
        cr = csv.DictReader(file)
        for r in cr:
            if r['Category'] and r['Sport Class'] and r['Factor']:
                cat = _cleankey(r['Category'])
                if cat != r['Category']:
                    _log.warning('Skipped malformed cat %s', r['Category'])
                    continue
                if cat not in self._store:
                    _log.info('Skipped unknown cat %s', cat)
                    continue
                sportclass = _cleankey(r['Sport Class'])
                if sportclass != r['Sport Class']:
                    _log.warning('Skipped malformed sport class %s',
                                 r['Sport Class'])
                    continue
                flbl = '.'.join((cat, sportclass))
                factor = readFactor(r['Factor'], label=flbl)
                if factor is not None:
                    _log.debug('%s add factor %s', flbl, factor)
                    self._store[cat][sportclass] = factor
                    count += 1
        if count:
            _log.debug('Loaded %d factors', count)
            self._valid = True
        else:
            _log.debug('No valid factors loaded')
            self._valid = False

    def load(self, filename=None):
        """Load time factors from filename."""
        if filename is None:
            filename = _DEFAULT_FILE
        srcfile = default_file(filename)
        if not os.path.exists(srcfile):
            _log.warning('Source file not found')
            return self._valid
        try:
            with open(srcfile) as f:
                self.read(f)
        except Exception as e:
            _log.warning('%s load: %s', e.__class__.__name__, e)
        return self._valid



# SPDX-License-Identifier: MIT
"""Telegraph

 Telegraph is a thin wrapper over paho.mqtt.client, which provides
 a MQTT pub/sub interface altered for use with metarace applications.

 Example:

	import metarace
	from metarace.telegraph import telegraph, from_json
	from time import sleep, time
	
	metarace.init()  # load shared configuration
	
	def messagecb(topic, message):
	    obj = from_json(message)
	    print(repr(obj))
	
	t = telegraph()
	t.set_will_json(obj={'example':[time(), None, None]},
	                topic='topic', qos=2, retain=True)
	t.subscribe('thetopic')
	t.setcb(messagecb)
	t.start()
	while not t.connected():  # wait for connection to broker
	    sleep(0)
	# ...
	t.publish_json(obj={'example':[time(),2,3]},
	               topic='topic', qos=2, retain=True)
	# ...
	t.flush()  # wait for any queued messages to be sent
	t.exit()  # request shutdown
	t.join()  # wait for clean shutdown

 Message callback functions receive two named parameters 'topic' and
 'message' which are both unicode strings. The message callback is run in
 the telegraph thread context. Use the convenience function "from_json"
 to convert a message from json into a python object. See defcallback
 for an example.

 Configuration is via metarace system config (metarace.json), under
 section 'telegraph':

  key: (type) Description [default]
  --
  host : (string) MQTT broker, None to disable ['localhost']
  port: (int) MQTT port [1883/8883]
  username : (string) username [None]
  password : (string) password [None]
  usetls : (bool) if True, connect to server over TLS [False]
  persist : (bool) if true, open a persistent connection to broker [False]
  clientid : (string) provide an explicit client id [None]
  qos : (int) default QOS to use for subscribe and publish [0]
  debug : (bool) if True, enable logging in MQTT library [False]
  deftopic : (string) a default publish topic [None]


"""

import threading
import queue
import logging
import json
import paho.mqtt.client as mqtt
from paho.mqtt.enums import CallbackAPIVersion
from uuid import uuid4
import metarace

_QUEUE_TIMEOUT = 2
_FLUSH_TIMEOUT = 30  # Maximum time to wait for unsent messages

# module logger
_log = logging.getLogger('telegraph')
_log.setLevel(logging.DEBUG)

_CONFIG_SCHEMA = {
    'ttype': {
        'prompt': 'MQTT Broker Details',
        'control': 'section',
    },
    'host': {
        'prompt': 'Host:',
        'hint': 'Hostname or IP of MQTT broker',
        'default': 'localhost',
        'defer': True,
    },
    'port': {
        'prompt': 'Port:',
        'control': 'short',
        'hint': 'TCP port number of MQTT broker',
        'type': 'int',
        'defer': True,
    },
    'usetls': {
        'prompt': 'Security:',
        'subtext': 'Use TLS?',
        'type': 'bool',
        'control': 'check',
        'hint': 'Connect to MQTT broker using TLS',
        'default': False,
        'defer': True,
    },
    'username': {
        'prompt': 'Username:',
        'hint': 'Username on MQTT broker if required',
        'defer': True,
    },
    'password': {
        'prompt': 'Password:',
        'hint': 'Password on MQTT broker if required',
        'defer': True,
    },
    'debug': {
        'prompt': 'Log:',
        'control': 'check',
        'type': 'bool',
        'subtext': 'Debug MQTT connection?',
        'hint': 'Log detailed information on connection to MQTT broken',
        'default': False,
        'defer': True,
    },
    'ssec': {
        'prompt': 'Session Options',
        'control': 'section',
    },
    'clientid': {
        'prompt': 'Client ID:',
        'hint': 'Specify client id for MQTT connection',
        'defer': True,
    },
    'persist': {
        'prompt': 'Session:',
        'subtext': 'Persistent?',
        'type': 'bool',
        'control': 'check',
        'hint': 'Request persistent session on broker',
        'default': True,
        'defer': True,
    },
    'qos': {
        'prompt': 'QoS:',
        'control': 'choice',
        'hint': 'Default message QOS',
        'type': 'int',
        'options': {
            '0': '0 - At most once',
            '1': '1 - At least once',
            '2': '2 - Exactly once',
        },
        'default': 0,
        'defer': True,
    },
    'deftopic': {
        'prompt': 'Default Topic:',
        'hint': 'Default topic for messages published without one',
        'defer': True,
    },
}


def from_json(payload=None):
    """Return message payload decoded from json, or None."""
    ret = None
    try:
        ret = json.loads(payload)
    except Exception as e:
        _log.warning('%s decoding JSON payload: %s', e.__class__.__name__, e)
    return ret


def defcallback(topic=None, message=None):
    """Default message receive callback function."""
    ob = from_json(message)
    if ob is not None:
        _log.debug('RCV %r: %r', topic, ob)
    else:
        _log.debug('RCV %r: %r', topic, message)


class telegraph(threading.Thread):
    """Metarace telegraph server thread."""

    def subscribe(self, topic=None, qos=None):
        """Add topic to the set of subscriptions with optional qos."""
        if topic:
            self.__subscriptions[topic] = qos
            if self.__connected:
                self.__queue.put_nowait(('SUBSCRIBE', topic, qos))

    def unsubscribe(self, topic=None):
        """Remove topic from the set of subscriptions."""
        if topic and topic in self.__subscriptions:
            del self.__subscriptions[topic]
            if self.__connected:
                self.__queue.put_nowait(('UNSUBSCRIBE', topic))

    def unsubscribe_all(self):
        """Remove all topics from the set of subscriptions."""
        topics = self.__subscriptions
        self.__subscriptions = {}
        if self.__connected:
            for topic in topics:
                self.__queue.put_nowait(('UNSUBSCRIBE', topic))

    def get_subscriptions(self):
        """Return a dictionary of current subscriptions"""
        ret = {}
        for topic, qos in self.__subscriptions.items():
            ret[topic] = qos
        return ret

    def setcb(self, func=None):
        """Set the message receive callback function."""
        if func is not None:
            self.__cb = func
        else:
            self.__cb = defcallback

    def set_deftopic(self, topic=None):
        """Set or clear the default publish topic."""
        if isinstance(topic, str) and topic:
            self.__deftopic = topic
        else:
            self.__deftopic = None
        _log.debug('Default publish topic set to: %r', self.__deftopic)

    def set_will_json(self,
                      obj=None,
                      topic=None,
                      qos=None,
                      retain=False,
                      cls=None,
                      indent=None):
        """Pack the provided object into JSON and set as will."""
        try:
            self.set_will(json.dumps(obj, cls=cls, indent=indent), topic, qos,
                          retain)
        except Exception as e:
            _log.error('Error setting will object %r: %s', obj, e)

    def set_will(self, message=None, topic=None, qos=None, retain=False):
        """Set or clear the last will with the broker."""
        if not self.__connect_pending and not self.__connected:
            if topic is not None:
                nqos = qos
                if nqos is None:
                    nqos = self.__qos
                payload = None
                if message is not None:
                    payload = message.encode('utf-8')
                self.__client.will_set(topic, payload, nqos, retain)
                _log.debug('Will set on topic %r', topic)
            else:
                self.__client.will_clear()
                _log.debug('Cleared will')
        else:
            _log.error('Unable to set will, already connected')

    def connected(self):
        """Return true if connected."""
        return self.__connected

    def flush(self):
        """Convenience function for clearing unsent messages"""
        self.__queue.put(('PUBLISH', ' ', '', 2, False, _FLUSH_TIMEOUT))
        self.__queue.put(('FLUSH', None))
        self.wait()

    def reconnect(self):
        """Request re-connection to broker."""
        self.__queue.put_nowait(('RECONNECT', True))

    def exit(self, msg=None):
        """Request thread termination."""
        self.__running = False
        self.__queue.put_nowait(('EXIT', msg))

    def wait(self):
        """Suspend calling thread until command queue is processed."""
        self.__queue.join()

    def publish(self,
                message=None,
                topic=None,
                qos=None,
                retain=False,
                timeout=None):
        """Queue provided message for publishing to nominated topic.

        If timeout (float seconds > 0) is provided, telegraph
        will wait up to timeout seconds after publishing for
        completion before re-processing the command queue.

        For QoS == 0, timeout applies to message delivery
        For QoS == 1, timeout applies to PUBACK
        For QoS == 2, timeout applies to PUBCOMP

        This method returns immediately even if timeout set.
        """
        self.__queue.put_nowait(
            ('PUBLISH', topic, message, qos, retain, timeout))

    def publish_json(self,
                     obj=None,
                     topic=None,
                     qos=None,
                     retain=False,
                     cls=None,
                     indent=None,
                     timeout=None):
        """Pack the provided object into JSON and publish to topic."""
        try:
            self.publish(json.dumps(obj, cls=cls, indent=indent), topic, qos,
                         retain, timeout)
        except Exception as e:
            _log.error('Error publishing object %r: %s', obj, e)

    def __init__(self):
        """Constructor."""
        threading.Thread.__init__(self, daemon=True)
        self.__queue = queue.Queue()
        self.__cb = defcallback
        self.__subscriptions = {}
        self.__connected = False
        self.__connect_pending = False
        self.__cid = None
        self.__resub = True
        self.__doreconnect = False

        metarace.sysconf.add_section('telegraph', _CONFIG_SCHEMA)
        self.__host = metarace.sysconf.get_value('telegraph', 'host')
        self.__port = metarace.sysconf.get_value('telegraph', 'port')
        if self.__port is None:
            self.__port = 1883
        self.__deftopic = metarace.sysconf.get_value('telegraph', 'deftopic')
        self.__qos = metarace.sysconf.get_value('telegraph', 'qos')
        self.__persist = metarace.sysconf.get_value('telegraph', 'persist')
        self.__cid = metarace.sysconf.get_value('telegraph', 'clientid')

        # check values
        if self.__qos > 2:
            _log.info('Invalid QOS %r set to %r', self.__qos, 2)
            self.__qos = 2
        if not self.__cid:
            self.__cid = str(uuid4())
        _log.debug('Using QoS: %r', self.__qos)
        _log.debug('Using client id: %r', self.__cid)
        _log.debug('Persistent connection: %r', self.__persist)

        # create mqtt client
        self.__client = mqtt.Client(
            callback_api_version=CallbackAPIVersion.VERSION2,
            client_id=self.__cid,
            clean_session=not self.__persist)

        if metarace.sysconf.get_value('telegraph', 'debug'):
            _log.debug('Enabling mqtt/paho debug')
            mqlog = logging.getLogger('telegraph.mqtt')
            mqlog.setLevel(logging.DEBUG)
            self.__client.enable_logger(mqlog)

        if metarace.sysconf.get_value('telegraph', 'usetls'):
            _log.debug('Enabling TLS connection')
            if not metarace.sysconf.has_value('telegraph', 'port'):
                # Update port for TLS if not exlicitly set
                self.__port = 8883
                _log.debug('Set port to %r', self.__port)
            self.__client.tls_set()
        username = metarace.sysconf.get_value('telegraph', 'username')
        password = metarace.sysconf.get_value('telegraph', 'password')
        if username and password:
            self.__client.username_pw_set(username, password)

        self.__client.reconnect_delay_set(2, 16)
        self.__client.on_message = self.__on_message
        self.__client.on_connect = self.__on_connect
        self.__client.on_disconnect = self.__on_disconnect
        if self.__host:
            self.__doreconnect = True
        self.__running = False

    def __reconnect(self):
        if not self.__connect_pending:
            if self.__connected:
                _log.debug('Disconnecting client')
                self.__client.disconnect()
                self.__client.loop_stop()
            if self.__host:
                _log.debug('Connecting to %s:%d', self.__host, self.__port)
                self.__connect_pending = True
                self.__client.connect_async(self.__host, self.__port)
                self.__client.loop_start()

    # PAHO methods - Callback API Version=2
    def __on_connect(self, client, userdata, flags, reason_code, properties):
        if reason_code == 0:
            _log.debug('Connect %r: %r/%r', client._client_id, flags,
                       reason_code)
            if not self.__resub and self.__persist and flags.session_present:
                _log.debug('Resumed existing session for %r',
                           client._client_id)
            else:
                _log.debug('Assuming Clean session for %r', client._client_id)
                s = []
                for t in self.__subscriptions:
                    nqos = self.__subscriptions[t]
                    if nqos is None:
                        nqos = self.__qos
                    s.append((t, nqos))
                if len(s) > 0:
                    _log.debug('Subscribe topics: %r', s)
                    self.__client.subscribe(s)
                self.__resub = False
            self.__connected = True
        else:
            _log.info('Connect failed with error %r: %r', reason_code,
                      mqtt.connack_string(reason_code))
            self.__connected = False
        self.__connect_pending = False

    def __on_disconnect(self, client, userdata, flags, reason_code,
                        properties):
        _log.debug('Disconnect %r: %r', client._client_id, reason_code)
        self.__connected = False
        # Note: PAHO lib will attempt re-connection automatically

    def __on_message(self, client, userdata, message):
        #_log.debug('Message from %r: %r', client._client_id, message)
        self.__cb(topic=message.topic, message=message.payload.decode('utf-8'))

    def run(self):
        """Called via threading.Thread.start()."""
        self.__running = True
        if self.__host:
            _log.debug('Starting')
        else:
            _log.debug('Not connected')
        while self.__running:
            try:
                # Check connection status
                if self.__host and self.__doreconnect:
                    self.__doreconnect = False
                    if not self.__connect_pending:
                        self.__reconnect()
                # Process command queue
                while self.__running:
                    m = self.__queue.get(timeout=_QUEUE_TIMEOUT)
                    self.__queue.task_done()
                    if m[0] == 'PUBLISH':
                        ntopic = self.__deftopic
                        if m[1] is not None:  # topic is set
                            ntopic = m[1]
                        nqos = m[3]
                        if nqos is None:
                            nqos = self.__qos
                        if ntopic:
                            msg = None
                            if m[2] is not None:
                                msg = m[2].encode('utf-8')

                            mi = self.__client.publish(ntopic, msg, nqos, m[4])

                            # Wait for publish if timeout provided
                            if m[5] is not None and m[5] > 0:
                                if mi.rc == 0:
                                    mi.wait_for_publish(m[5])
                                else:
                                    if nqos == 0:
                                        _log.warning(
                                            'Publish to %r dropped: %r,%r',
                                            ntopic, mi.rc, mi.mid)
                    elif m[0] == 'SUBSCRIBE':
                        _log.debug('Subscribe topic: %r q=%r', m[1], m[2])
                        nqos = m[2]
                        if nqos is None:
                            nqos = self.__qos
                        self.__client.subscribe(m[1], nqos)
                    elif m[0] == 'UNSUBSCRIBE':
                        _log.debug('Un-subscribe topic: %r', m[1])
                        self.__client.unsubscribe(m[1])
                    elif m[0] == 'RECONNECT':
                        self.__connect_pending = False
                        self.__doreconnect = True
                    elif m[0] == 'EXIT':
                        _log.debug('Request to close: %r', m[1])
                        self.__running = False
            except queue.Empty:
                pass
            except Exception as e:
                _log.error('%s: %s', e.__class__.__name__, e)
                self.__connect_pending = False
                self.__doreconnect = False
        if self.__connected:
            self.__client.disconnect()
            self.__client.loop_stop()
        _log.info('Exiting')
